from autokeras.tasks.image import ImageClassifier
from autokeras.tasks.image import ImageRegressor
from autokeras.tasks.structured_data import StructuredDataClassifier
from autokeras.tasks.structured_data import StructuredDataRegressor
from autokeras.tasks.text import TextClassifier
from autokeras.tasks.text import TextRegressor
from autokeras.tasks.time_series_forecaster import TimeseriesForecaster
